library(readxl)

set.seed(11)

###############
## read data ##
###############

project.dir<-"RealDataAnalysis"

dir<-file.path(project.dir,"data")

case_age<-as.matrix(read_excel(file.path(dir,"case_age1.xls"),col_names=FALSE))

case_cho<-as.matrix(read_excel(file.path(dir,"case_cholester1.xls"),col_names=FALSE))
case_dia<-as.matrix(read_excel(file.path(dir,"case_diastolic1.xls"),col_names=FALSE))
case_sys<-as.matrix(read_excel(file.path(dir,"case_systolic1.xls"),col_names=FALSE))

control_age<-as.matrix(read_excel(file.path(dir,"control_age1.xls"),col_names=FALSE))

control_cho<-as.matrix(read_excel(file.path(dir,"control_cholester1.xls"),col_names=FALSE))
control_dia<-as.matrix(read_excel(file.path(dir,"control_diastolic1.xls"),col_names=FALSE))
control_sys<-as.matrix(read_excel(file.path(dir,"control_systolic1.xls"),col_names=FALSE))

ncase<-nrow(case_age)
ncontrol<-nrow(control_age)
nind<-ncase+ncontrol
nobs<-7

ttij_raw<-matrix(NA,nind,nobs)
xxij_raw<-array(NA,dim=c(nind,nobs,3))

ttij_raw[1:ncase,]<-case_age
ttij_raw[ncase+1:ncontrol,]<-control_age

xxij_raw[1:ncase,,1]<-case_cho
xxij_raw[1:ncase,,2]<-case_dia
xxij_raw[1:ncase,,3]<-case_sys

xxij_raw[ncase+1:ncontrol,,1]<-control_cho
xxij_raw[ncase+1:ncontrol,,2]<-control_dia
xxij_raw[ncase+1:ncontrol,,3]<-control_sys

standardizing<-c(sd(xxij_raw[,,1]),sd(xxij_raw[,,2]),sd(xxij_raw[,,3]))

xxij_raw[,,1]<-(xxij_raw[,,1]-mean(xxij_raw[,,1]))/sd(xxij_raw[,,1])
xxij_raw[,,2]<-(xxij_raw[,,2]-mean(xxij_raw[,,2]))/sd(xxij_raw[,,2])
xxij_raw[,,3]<-(xxij_raw[,,3]-mean(xxij_raw[,,3]))/sd(xxij_raw[,,3])

ot_raw<-as.vector(ttij_raw[,7])
st_raw<-as.vector(ttij_raw[,1])
delta_raw<-rep(c(TRUE,FALSE),c(ncase,ncontrol))

#############
# partition #
#############

testcase<-1:ncase%in%sample(1:ncase,ncase/3)
testcontrol<-1:ncontrol%in%sample(1:ncontrol,ncontrol/3)
testlabel<-c(testcase,testcontrol)
trainlabel<-!testlabel
sum(testlabel);sum(trainlabel)

alltimepoints<-seq(min(ttij_raw),max(ttij_raw))

ncase_train<-sum(!testcase)
ncase_test<-sum(testcase)

xxij_train<-xxij_raw[trainlabel,,]
ttij_train<-ttij_raw[trainlabel,]
ot_train<-ot_raw[trainlabel]
st_train<-st_raw[trainlabel]
delta_train<-delta_raw[trainlabel]

xxij_test<-xxij_raw[!trainlabel,,]
ttij_test<-ttij_raw[!trainlabel,]
ot_test<-ot_raw[!trainlabel]
st_test<-st_raw[!trainlabel]
delta_test<-delta_raw[!trainlabel]

#############
## CV beta ##
#############

negloglik_leave<-function(beta){
  negloglik_leave<-0
  rrij<-apply(xxij_train,c(1,2),function(x)x%*%beta)
  for(ii in 1:ncase_train){
    if(ii==iileave)next
    negloglik_leave<-negloglik_leave-rrij[ii,7]
    Ti<-ot_train[ii]
    riskset<-ot_train>=Ti
    riskset[iileave]<-FALSE
    k_term<-0.75*(1-((Ti-ttij_train[riskset,])/hh_int)^2)/hh_int
    k_term[k_term<0]<-0
    log_term<-log(sum(k_term*exp(rrij[riskset,])))
    negloglik_leave<-negloglik_leave+log_term
  }
  return(negloglik_leave)
}


hhchoice<-seq(6,20,2)
calculate_CVL<-numeric(length(hhchoice))
for(hh_index in 1:length(hhchoice)){
  hh_int<-hhchoice[hh_index]
  allkvalues<-0.75*(1-(-hh_int:hh_int/hh_int)^2)
  for(iileave in 1:nrow(xxij_train)){
    cat(iileave)
    beta_leave<-optim(c(0,0,0,0),negloglik_leave,"L-BFGS-B")$par
    rrij_leave<-matrix(NA,nrow=nrow(xxij_train),ncol=ncol(xxij_train))
    for(ii in 1:nrow(xxij_train)){
      for(jj in 1:ncol(xxij_train)){
        rrij_leave[ii,jj]<-xxij_train[ii,jj,]%*%beta_leave
      }
    }
    lliscore<-delta_train[iileave]
    for(ii in 1:ncase_train){
      if(ot_train[ii]>ot_train[iileave])next
      numerator<-0;denominator<-0
      for(jjleave in 1:7){
        if(abs(ot_train[ii]-ttij_train[iileave,jjleave])>=hh_int)next
        numerator<-numerator+allkvalues[ot_train[ii]-ttij_train[iileave,jjleave]+hh_int+1]*exp(rrij_leave[iileave,jjleave])
      }
      for(ll in 1:nrow(xxij_train)){
        if(ot_train[ll]>ot_train[ii])next
        for(jj in 1:7){
          if(abs(ttij_train[ll,jj]-ot_train[ii])>=hh_int)next
          denominator<-denominator+allkvalues[ot_train[ii]-ttij_train[ll,jj]+hh_int+1]*exp(rrij_leave[ll,jj])
        }
      }
      lliscore<-lliscore-numerator/denominator
    }
    calculate_CVL[hh_index]<-calculate_CVL[hh_index]+lliscore^2
  }
  print(c("CVL",calculate_CVL))
}
print(calculate_CVL)

hh_int<-hhchoice[which.min(calculate_CVL)]
hh_beta<-hh_int

##########
## beta ##
##########

negloglik<-function(beta){
  negloglik<-0
  rrij<-apply(xxij_train,c(1,2),function(x)x%*%beta)
  for(ii in 1:ncase_train){
    negloglik<-negloglik-rrij[ii,7]
    Ti<-ot_train[ii]
    riskset<-ot_train>=Ti
    k_term<-0.75*(1-((Ti-ttij_train[riskset,])/hh_int)^2)/hh_int
    k_term[k_term<0]<-0
    log_term<-log(sum(k_term*exp(rrij[riskset,])))
    negloglik<-negloglik+log_term
  }
  return(negloglik)
}

beta_est<-optim(c(0,0,0),negloglik,"L-BFGS-B")$par
show(beta_est)
round(beta_est/standardizing,5)

rrij_train<-matrix(NA,nrow=nrow(xxij_train),ncol=ncol(xxij_train))
for(ii in 1:nrow(xxij_train)){
  for(jj in 1:ncol(xxij_train)){
    rrij_train[ii,jj]<-xxij_train[ii,jj,]%*%beta_est
  }
}

########
## mu ##
########

library(Rcpp)
rcpp_dir<-file.path(project.dir,"rcpp")
sourceCpp(file.path(rcpp_dir,"mean_est.cpp"))
sourceCpp(file.path(rcpp_dir,"mean_CV_faster.cpp"))

hh_vec<-seq(2,20,2)
CV<-mean_CV_faster(rrij_train,ttij_train-min(ttij_raw),hh_vec,72)
hh_int<-hh_vec[which.min(CV)]
hh_mu<-hh_int
mu<-mean_est(rrij_train,ttij_train-min(ttij_raw),hh_int,72)

eps2_train<-matrix(NA,nrow=nrow(rrij_train),ncol=ncol(rrij_train))
for(ii in 1:nrow(rrij_train)){
  eps2_train[ii,]<-(rrij_train[ii,]-mu[ttij_train[ii,]-alltimepoints[1]+1])^2
}

hh_vec<-seq(2,20,2)
CV<-mean_CV_faster(eps2_train,ttij_train-min(ttij_raw),hh_vec,72)
hh_int<-hh_vec[which.min(CV)]
hh_sigma<-hh_int
sigma2<-mean_est(eps2_train,ttij_train-min(ttij_raw),hh_int,72)
plot(sigma2)

rrstd_train<-matrix(NA,nrow=nrow(rrij_train),ncol=ncol(rrij_train))
for(ii in 1:nrow(rrij_train)){
  rrstd_train[ii,]<-(rrij_train[ii,]-mu[ttij_train[ii,]-alltimepoints[1]+1])/
    sqrt(sigma2[ttij_train[ii,]-alltimepoints[1]+1])
}

###################

interval<-c()
for(ii in 1:nrow(xxij_train)){
  for(jj in 2:7){
    interval<-c(interval,ttij_train[ii,jj]-ttij_train[ii,jj-1])
  }
}
q<-mean(interval)

lambda<-0.1
EEij_train<-matrix(NA,nrow=nrow(rrstd_train),ncol=ncol(rrstd_train))
for(ii in 1:nrow(xxij_train)){
  V<-1-(1-lambda)^q
  E<-V*rrstd_train[ii,1]
  EEij_train[ii,1]<-E
  for(jj in 2:7){
    V<-V/(V+(1-lambda)^((ttij_train[ii,jj]-ttij_train[ii,jj-1])))
    E<-(1-V)*E+V*rrstd_train[ii,jj]
    EEij_train[ii,jj]<-E
  }
}

ll_list<-seq(0,3,0.01)
ATS<-numeric(length(ll_list))
for(ll_index in 1:length(ll_list)){
  ll<-ll_list[ll_index]
  ind1<-1:nrow(EEij_train)
  ind2<-apply(EEij_train,1,function(x)which(x>ll)[1])
  ATS[ll_index]<-mean(ttij_train[cbind(ind1,ind2)]-st_train,na.rm=TRUE)
}

ll<-ll_list[which(ATS>10)[1]]
show(ATS)
show(ll)

##########
## test ##
##########

rrij_test<-matrix(NA,nrow=nrow(xxij_test),ncol=ncol(xxij_test))
for(ii in 1:nrow(xxij_test)){
  for(jj in 1:ncol(xxij_test)){
    rrij_test[ii,jj]<-xxij_test[ii,jj,]%*%beta_est
  }
}

rrstd_test<-matrix(NA,nrow=nrow(rrij_test),ncol=ncol(rrij_test))

for(ii in 1:nrow(rrstd_test)){
  rrstd_test[ii,]<-(rrij_test[ii,]-mu[ttij_test[ii,]-alltimepoints[1]+1])/
    sigma2[ttij_test[ii,]-alltimepoints[1]+1]
}

lambda<-0.1
EEij_test<-matrix(NA,nrow=nrow(rrstd_test),ncol=ncol(rrstd_test))
for(ii in 1:nrow(EEij_test)){
  V<-1-(1-lambda)^q
  E<-V*rrstd_test[ii,1]
  EEij_test[ii,1]<-E
  for(jj in 2:7){
    V<-V/(V+(1-lambda)^((ttij_test[ii,jj]-ttij_test[ii,jj-1])))
    E<-(1-V)*E+V*rrstd_test[ii,jj]
    EEij_test[ii,jj]<-E
  }
}

ind2_test<-apply(EEij_test,1,function(x)which(x>ll)[1])
sum(!is.na(ind2_test)[1:ncase_test])
sum(!is.na(ind2_test)[-(1:ncase_test)])
sum(is.na(ind2_test)[-(1:ncase_test)])

E<-as.vector(t(EEij_test[1:ncase_test,]))
j<-rep(1:7,ncase_test)
id<-paste(rep(1:ncase_test,each=7))
id<-factor(id,paste(1:27))
EWMA_df<-data.frame(E=E,j=j,id=id)