program main_program

implicit none

integer,parameter::nrep=1000,nind=500,ntimepoints=1000,dimbeta=3,nsim=100,nmax_obs=100
integer,parameter::partition=1
integer,parameter::rep_start=50*(partition-1)+1,rep_end=50*partition
double precision,parameter::omega=1.0/ntimepoints,pi=3.141592653589793d0,rate=20d0
integer::ii,jj,kk,ll,mm,tt,rep,writerep
double precision,parameter,dimension(ntimepoints)::alltimepoints=(/(ii*omega,ii=1,ntimepoints)/),hazard0=0.2!alltimepoints/2d0
double precision,parameter::beta_true(dimbeta)=[0.5,0.3,0.2]
double precision::beta_old(dimbeta),beta_new(dimbeta),beta_est(dimbeta)
double precision::xx_true(nind,ntimepoints,dimbeta),xxij(nind,nmax_obs,dimbeta),xx_temp(ntimepoints,dimbeta)
double precision::ttij(nind,nmax_obs),rrij(nind,nmax_obs),rrij_std(nind,nmax_obs)
double precision::epsij(nind,nmax_obs),epsij2(nind,nmax_obs)
double precision::ft(nind),ct(nind),ot(nind)
integer::ttij_int(nind,nmax_obs),ttij_temp(nmax_obs)
integer::ft_int(nind),ct_int(nind),ot_int(nind),mmi(nind)
logical::delta(nind)
double precision,dimension(ntimepoints)::mu_est,sigma2_est,sigma_est
double precision,dimension(ntimepoints)::lambda,tr
double precision::randexp
double precision,parameter::condition_value=1e-8

integer,parameter::nhh_beta=10,nhh_mu=10,nhh_sigma2=10
integer::hh_int
double precision::hh,CVL(nhh_beta),CVmu(nhh_mu),CVsigma2(nhh_sigma2)
integer::hhtriplet(3)

!! ll lambda
integer::lambda_index
integer,parameter::ncontrol_lambda=4
double precision,parameter::control_lambda(ncontrol_lambda)=[0.01,0.02,0.03,0.05]

!! monitor
integer,parameter::nind_OC=10000
double precision::xxij_OC(nind_OC,nmax_obs,dimbeta),rrij_OC(nind_OC,nmax_obs),rrij_std_OC(nind_OC,nmax_obs)
integer,dimension(nind_OC)::ct_int_OC,ot_int_OC,ft_int_OC
integer::ttij_int_OC(nind_OC,nmax_obs),mmi_OC(nind_OC)
logical::delta_OC(nind_OC)
double precision::xx_true_OC(ntimepoints,dimbeta)

integer,parameter::nshift=41
integer::shift_index
double precision,parameter::shift_list(nshift)=(/(ii*0.2-0.2,ii=1,nshift)/)
double precision::shift

double precision::temp_ATS
integer::temp_table(4)

double precision::all_ATS(nrep,nshift,ncontrol_lambda+1)
double precision::control_ll(ncontrol_lambda+1)
double precision::all_ll(nrep,ncontrol_lambda+1)
integer::all_table(nrep,nshift,4,ncontrol_lambda+1)

double precision::all_beta(nrep,dimbeta)
integer::all_hh_int(nrep,3)

integer,parameter::nseed=33
double precision::seed_matrix(nrep,nseed)
double precision::start,finish

call random_seed(put=(/(12345*ii,ii=1,nseed)/))
call random_number(seed_matrix)

write(6,*)"start",rep_start,"end",rep_end
do rep=rep_start,rep_end
write(6,*)"rep=",rep
call flush(6)

call random_seed(put=floor(100000*seed_matrix(rep,:)))

! PROGRAM 1 - GENERATE DATA

xxij=-10
rrij=-10
do ii=1,nind
 ct_int(ii)=minval([nint(random_Weibull(10d0)*ntimepoints),1000])
 ct(ii)=ct_int(ii)*omega
 xx_temp=generate_xx()
 xx_true(ii,:,1)=xx_temp(:,1)
 xx_true(ii,:,2)=xx_temp(:,2)
 xx_true(ii,:,3)=xx_temp(:,3)
 lambda=hazard0*exp(xx_true(ii,:,1)*beta_true(1)+&
                    xx_true(ii,:,2)*beta_true(2)+&
                    xx_true(ii,:,3)*beta_true(3))
 tr=cumsum(lambda)*omega
 randexp=random_exponential()
 ft_int(ii)=count(randexp>tr)+1
 ft(ii)=ft_int(ii)*omega
 ot_int(ii)=minval([ft_int(ii),ct_int(ii)])
 ot(ii)=minval([ft(ii),ct(ii)])
 delta(ii)=ft_int(ii)<=ct_int(ii)
 do
  ttij_temp=simulate_path(rate)
  if(ttij_temp(nmax_obs-1)>=ot_int(ii))exit
 end do
 mmi(ii)=count(ttij_temp<ot_int(ii))
 if(mmi(ii)==0)then
  ttij_int(ii,1)=ot_int(ii)
  ttij_int(ii,2:nmax_obs)=-1!NaN
 else if(mmi(ii)==nmax_obs-1)then
  ttij_int(ii,1:nmax_obs-1)=ttij_temp(1:nmax_obs-1)
  ttij_int(ii,nmax_obs)=ot_int(ii)
 else
  ttij_int(ii,1:mmi(ii))=ttij_temp(1:mmi(ii))
  ttij_int(ii,mmi(ii)+1)=ot_int(ii)
  ttij_int(ii,mmi(ii)+2:nmax_obs)=-1!NaN
 end if
 mmi(ii)=mmi(ii)+1
 do kk=1,dimbeta
  xxij(ii,1:mmi(ii),kk)=xx_true(ii,ttij_int(ii,1:mmi(ii)),kk)
 end do
end do
ttij=ttij_int*omega

call sort_by_time

! PROGRAM 2 - WRITE DATA TO FILES

!do kk=1,dimbeta
! open(5,file="outputdata/rep"//int_to_char(rep)//"xxij"//int_to_char(kk)//".dat")
! do ii=1,nind
!  write(5,"(1000(f16.12))")(xxij(ii,jj,kk),jj=1,nmax_obs)
! end do
! close(5)
!end do

!open(5,file="outputdata/rep"//int_to_char(rep)//"ttij.dat")
!do ii=1,nind
! write(5,"(1000(I7))")(ttij_int(ii,jj),jj=1,nmax_obs)
!end do
!close(5)

!open(5,file="outputdata/rep"//int_to_char(rep)//"mmi.dat")
!write(5,"(1000(I7))")mmi
!close(5)

! PROGRAM 3 - NEWTON-RAPHSON

hh_int=100

call cpu_time(start)
CVL=calculate_CVL(hh_int) !hh_int=100
call cpu_time(finish)

beta_est=estimate_beta(hh_int)
all_beta(rep,:)=beta_est
all_hh_int(rep,1)=hh_int
print*,"estimated beta:",beta_est
write(6,*)"selected bandwidth for beta",hh_int,"time",finish-start,"s"

do ii=1,nind
do jj=1,mmi(ii)
rrij(ii,jj)=dot_product(xxij(ii,jj,:),beta_est)
end do
end do

call cpu_time(start)
CVmu=calculate_CVmu_faster(hh_int)  !hh_int=100
call cpu_time(finish)

mu_est=estimate_mu(hh_int)
all_hh_int(rep,2)=hh_int
write(6,*)"selected bandwidth for mu",hh_int,"time",finish-start,"s"

do ii=1,nind
epsij(ii,1:mmi(ii))=rrij(ii,1:mmi(ii))-mu_est(ttij_int(ii,1:mmi(ii)))
end do
epsij2=epsij*epsij

call cpu_time(start)
CVsigma2=calculate_CVsigma2_faster(hh_int) !hh_int=100
call cpu_time(finish)

sigma2_est=estimate_sigma2(hh_int)
sigma_est=sqrt(sigma2_est)
all_hh_int(rep,3)=hh_int
write(6,*)"selected bandwidth for sigma2",hh_int,"time",hh_int,finish-start,"s"

rrij_std=10
do ii=1,nind
rrij_std(ii,1:mmi(ii))=(rrij(ii,1:mmi(ii))-mu_est(ttij_int(ii,1:mmi(ii))))/sigma_est(ttij_int(ii,1:mmi(ii)))
end do

! samples for calculating control limits

xxij=-10
rrij=-10
do ii=1,nind
 ct_int(ii)=minval([nint(random_Weibull(10d0)*ntimepoints),1000])
 ct(ii)=ct_int(ii)*omega
 xx_temp=generate_xx()
 xx_true(ii,:,1)=xx_temp(:,1)
 xx_true(ii,:,2)=xx_temp(:,2)
 xx_true(ii,:,3)=xx_temp(:,3)
 lambda=hazard0*exp(xx_true(ii,:,1)*beta_true(1)+&
                    xx_true(ii,:,2)*beta_true(2)+&
                    xx_true(ii,:,3)*beta_true(3))
 tr=cumsum(lambda)*omega
 randexp=random_exponential()
 ft_int(ii)=count(randexp>tr)+1
 ft(ii)=ft_int(ii)*omega
 ot_int(ii)=minval([ft_int(ii),ct_int(ii)])
 ot(ii)=minval([ft(ii),ct(ii)])
 delta(ii)=ft_int(ii)<=ct_int(ii)
 do
  ttij_temp=simulate_path(rate)
  if(ttij_temp(nmax_obs-1)>=ot_int(ii))exit
 end do
 mmi(ii)=count(ttij_temp<ot_int(ii))
 if(mmi(ii)==0)then
  ttij_int(ii,1)=ot_int(ii)
  ttij_int(ii,2:nmax_obs)=-1!NaN
 else if(mmi(ii)==nmax_obs-1)then
  ttij_int(ii,1:nmax_obs-1)=ttij_temp(1:nmax_obs-1)
  ttij_int(ii,nmax_obs)=ot_int(ii)
 else
  ttij_int(ii,1:mmi(ii))=ttij_temp(1:mmi(ii))
  ttij_int(ii,mmi(ii)+1)=ot_int(ii)
  ttij_int(ii,mmi(ii)+2:nmax_obs)=-1!NaN
 end if
 mmi(ii)=mmi(ii)+1
 do kk=1,dimbeta
  xxij(ii,1:mmi(ii),kk)=xx_true(ii,ttij_int(ii,1:mmi(ii)),kk)
 end do
end do
ttij=ttij_int*omega

do ii=1,nind
do jj=1,mmi(ii)
rrij(ii,jj)=dot_product(xxij(ii,jj,:),beta_est)
end do
end do

rrij_std=10
do ii=1,nind
rrij_std(ii,1:mmi(ii))=(rrij(ii,1:mmi(ii))-mu_est(ttij_int(ii,1:mmi(ii))))/sigma_est(ttij_int(ii,1:mmi(ii)))
end do

do lambda_index=1,ncontrol_lambda
control_ll(lambda_index)=choose_ll(control_lambda(lambda_index))
all_ll(rep,lambda_index)=control_ll(lambda_index)
end do

control_ll(ncontrol_lambda+1)=choose_ll_d()
all_ll(rep,ncontrol_lambda+1)=control_ll(ncontrol_lambda+1)

write(6,*)control_ll

! PROGRAM 6 - OUTPUT ESTIMATES

!open(5,file="output/rep"//int_to_char(rep)&
!//"n"//int_to_char(nind)//"mu.dat")
!write(5,"(1000(f16.12))")mu_est
!close(5)

!open(5,file="output/rep"//int_to_char(rep)&
!//"n"//int_to_char(nind)//"sigma2.dat")
!write(5,"(1000(f16.12))")sigma2_est
!close(5)

! MONITOR

do shift_index=1,nshift
shift=shift_list(shift_index)
xxij_OC=-10
rrij_OC=-10
do ii=1,nind_OC
 ct_int_OC(ii)=minval([nint(random_Weibull(10d0)*ntimepoints),1000])
 xx_temp=generate_xx()
 xx_true_OC(:,1)=beta_true(1)*shift+xx_temp(:,1)
 xx_true_OC(:,2)=beta_true(2)*shift+xx_temp(:,2)
 xx_true_OC(:,3)=beta_true(3)*shift+xx_temp(:,3)
 lambda=hazard0*exp(xx_true_OC(:,1)*beta_true(1)+&
                    xx_true_OC(:,2)*beta_true(2)+&
                    xx_true_OC(:,3)*beta_true(3))
 tr=cumsum(lambda)*omega
 randexp=random_exponential()
 ft_int_OC(ii)=count(randexp>tr)+1
 ot_int_OC(ii)=minval([ft_int_OC(ii),ct_int_OC(ii)])
 delta_OC(ii)=ft_int_OC(ii)<=ct_int_OC(ii)
 do
  ttij_temp=simulate_path(rate)
  if(ttij_temp(nmax_obs-1)>=ot_int_OC(ii))exit
 end do
 mmi_OC(ii)=count(ttij_temp<ot_int_OC(ii))
 if(mmi_OC(ii)==0)then
  ttij_int_OC(ii,1)=ot_int_OC(ii)
  ttij_int_OC(ii,2:nmax_obs)=-1!NaN
 else if(mmi_OC(ii)==nmax_obs-1)then
  ttij_int_OC(ii,1:nmax_obs-1)=ttij_temp(1:nmax_obs-1)
  ttij_int_OC(ii,nmax_obs)=ot_int_OC(ii)
 else
  ttij_int_OC(ii,1:mmi_OC(ii))=ttij_temp(1:mmi_OC(ii))
  ttij_int_OC(ii,mmi_OC(ii)+1)=ot_int_OC(ii)
  ttij_int_OC(ii,mmi_OC(ii)+2:nmax_obs)=-1!NaN
 end if
 mmi_OC(ii)=mmi_OC(ii)+1
 do kk=1,dimbeta
  xxij_OC(ii,1:mmi_OC(ii),kk)=xx_true_OC(ttij_int_OC(ii,1:mmi_OC(ii)),kk)
 end do
end do

do ii=1,nind_OC
do jj=1,mmi_OC(ii)
rrij_OC(ii,jj)=dot_product(xxij_OC(ii,jj,:),beta_est)
end do
end do

rrij_std_OC=10
do ii=1,nind_OC
rrij_std_OC(ii,1:mmi_OC(ii))=(rrij_OC(ii,1:mmi_OC(ii))-mu_est(ttij_int_OC(ii,1:mmi_OC(ii))))/&
    sigma_est(ttij_int_OC(ii,1:mmi_OC(ii)))
end do

do lambda_index=1,ncontrol_lambda
call monitor(control_ll(lambda_index),control_lambda(lambda_index),temp_ATS,temp_table)
all_ATS(rep,shift_index,lambda_index)=temp_ATS
all_table(rep,shift_index,1:4,lambda_index)=temp_table
end do
call monitor_d(control_ll(ncontrol_lambda+1),temp_ATS,temp_table)
all_ATS(rep,shift_index,ncontrol_lambda+1)=temp_ATS
all_table(rep,shift_index,1:4,ncontrol_lambda+1)=temp_table

end do

open(5,file="all_hh_int.dat")
do writerep=rep_start,rep_end
write(5,"(1000(I7))")all_hh_int(writerep,:)
end do
close(5)

open(5,file="all_beta.dat")
do writerep=rep_start,rep_end
write(5,"(1000(f16.12))")all_beta(writerep,:)
end do
close(5)

open(5,file="all_ATS.dat")
do writerep=rep_start,rep_end
write(5,"(1000(f16.02))")all_ATS(writerep,:,1),&
                         all_ATS(writerep,:,2),&
                         all_ATS(writerep,:,3),&
                         all_ATS(writerep,:,4),&
                         all_ATS(writerep,:,5)
end do
close(5)

open(5,file="all_ll.dat")
do writerep=rep_start,rep_end
write(5,"(1000(f16.10))")all_ll(writerep,:)
end do
close(5)

end do


contains

function random_normal()
double precision::r1,r2,random_normal
call random_number(r1)
call random_number(r2)
random_normal=sqrt(-2.0d0*log(r1))*cos(2.0d0*pi*r2)
end function random_normal

function random_exponential()
double precision::r,random_exponential
do
 call random_number(r)
 if(r>0.0d0)exit
end do
random_exponential=-log(r)
end function random_exponential

function rnorm(nn)
integer,intent(in)::nn
double precision::rnorm(nn)
integer::ii
do ii=1,nn
 rnorm(ii)=random_normal()
end do
end function

! SUBROUTINE 1 - RANDOM NUMBER GENERATORS

double precision FUNCTION random_Weibull(a)

!     Generates a random variate from the Weibull distribution with
!     probability density:
!                      a
!               a-1  -x
!     f(x) = a.x    e

double precision, INTENT(IN) :: a

!     For speed, there is no checking that a is not zero or very small.

random_Weibull = random_exponential() ** (1.0d0/a)
RETURN

END FUNCTION random_Weibull

function random_poisson(lambda)
integer::random_poisson
double precision,intent(in)::lambda
double precision::l,p,u
p=1d0;l=exp(-lambda)
random_poisson=0
do while(p>l)
 call random_number(u)
 random_poisson=random_poisson+1
 p=p*u
end do
random_poisson=random_poisson-1
end function random_poisson

function random_MVN3()
double precision::random_MVN3(3),Chol(3,3)
Chol=0.0
Chol(1,1)=1.0
Chol(2,1)=1.0/4.0;Chol(2,2)=sqrt(15.0/16.0)
                  Chol(3,2)=sqrt(1.0/15.0);Chol(3,3)=sqrt(14.0/15.0)
random_MVN3(1)=random_normal()
random_MVN3(2)=random_normal()
random_MVN3(3)=random_normal()
random_MVN3=matmul(Chol,random_MVN3)
end function

! SUBROUTINE 2 - UTILITY FUNCTIONS

function cumsum(xx)
double precision,dimension(:),intent(in)::xx
double precision,dimension(size(xx))::cumsum
integer::nn,jj
nn=size(xx)
cumsum(1)=xx(1)
do jj=2,nn
cumsum(jj)=cumsum(jj-1)+xx(jj)
end do
end function cumsum

function cumsum_int(xx)
integer,dimension(:),intent(in)::xx
integer,dimension(size(xx))::cumsum_int
integer::nn,jj
nn=size(xx)
cumsum_int(1)=xx(1)
do jj=2,nn
cumsum_int(jj)=cumsum_int(jj-1)+xx(jj)
end do
end function cumsum_int

function simulate_path(lambda)
integer,dimension(nmax_obs)::simulate_path
double precision::lambda
integer::jj
simulate_path(1)=1
do jj=2,nmax_obs
 simulate_path(jj)=simulate_path(jj-1)+random_poisson(lambda)+1
end do
end function simulate_path

function solve3by3(AA,bb) result(SOLVEAb)
double precision::AA(dimbeta,dimbeta),bb(dimbeta),SOLVEAb(dimbeta)
double precision::COFACTOR(dimbeta,dimbeta),DETERMINANT

   COFACTOR(1,1) = +(AA(2,2)*AA(3,3)-AA(2,3)*AA(3,2))
   COFACTOR(1,2) = -(AA(2,1)*AA(3,3)-AA(2,3)*AA(3,1))
   COFACTOR(1,3) = +(AA(2,1)*AA(3,2)-AA(2,2)*AA(3,1))
   COFACTOR(2,1) = -(AA(1,2)*AA(3,3)-AA(1,3)*AA(3,2))
   COFACTOR(2,2) = +(AA(1,1)*AA(3,3)-AA(1,3)*AA(3,1))
   COFACTOR(2,3) = -(AA(1,1)*AA(3,2)-AA(1,2)*AA(3,1))
   COFACTOR(3,1) = +(AA(1,2)*AA(2,3)-AA(1,3)*AA(2,2))
   COFACTOR(3,2) = -(AA(1,1)*AA(2,3)-AA(1,3)*AA(2,1))
   COFACTOR(3,3) = +(AA(1,1)*AA(2,2)-AA(1,2)*AA(2,1))
   
   DETERMINANT=AA(3,1)*COFACTOR(3,1)+AA(3,2)*COFACTOR(3,2)+AA(3,3)*COFACTOR(3,3)
   SOLVEAb=matmul(COFACTOR,bb)/DETERMINANT

end function solve3by3

subroutine print_double_matrix(BBB)
double precision,dimension(:,:)::BBB
integer::ii
integer,dimension(:),allocatable::nn
nn=shape(BBB)
do ii=1,nn(1)
write(*,"(20f12.6)")BBB(ii,:)
end do
end subroutine

subroutine print_double_vector(BBB)
double precision,dimension(:)::BBB
write(*,"(20f12.6)")BBB(:)
end subroutine

function int_to_char(nn)
integer,intent(in)::nn
character(4)::int_to_char
write(int_to_char,"(I0.4)")nn
end function

!!!!!!!!!!!!!!!
!!  Model 1  !!
!!!!!!!!!!!!!!!

function generate_xx() result(xx)
double precision::xx(ntimepoints,dimbeta),zetai(dimbeta),tau=0.2
zetai=random_MVN3()
xx(:,1)=2*alltimepoints+zetai(1)*tau+(1-alltimepoints**2/2)*rnorm(ntimepoints)
xx(:,2)=-sin(2*pi*alltimepoints)+zetai(2)*tau+(1+0.2*sin(3*pi*alltimepoints))*rnorm(ntimepoints)
xx(:,3)=log(1+alltimepoints)+zetai(3)*tau+(1+0.2*cos(3*pi*alltimepoints))*rnorm(ntimepoints)!!!
end function

!!!!!!!!!!!!!!!
!!  Model 2  !!
!!!!!!!!!!!!!!!

!function rts(nn)
!integer,intent(in)::nn
!double precision::rts(nn)
!integer::ii
!rts(1)=0.5*random_normal()/sqrt(1-0.81)
!do ii=2,nn
! rts(ii)=0.9*rts(ii-1)+0.5*random_normal()
!end do
!end function

!function generate_xx() result(xx)
!double precision::xx(ntimepoints,dimbeta),zetai(dimbeta)
!integer::tt
!xx(1,1)=0.3*random_normal()
!xx(1,2)=0.3*random_normal()
!xx(1,3)=0.3*random_normal()
!do tt=2,ntimepoints
! xx(tt,1)=xx(tt-1,1)+omega*random_normal()
! xx(tt,2)=xx(tt-1,2)+omega*random_normal()
! xx(tt,3)=xx(tt-1,3)+omega*random_normal()
!end do
!xx(:,1)=sin(pi*alltimepoints)/2+xx(:,1)+(1-0.25*cos(3*pi*alltimepoints))*rts(ntimepoints)
!xx(:,2)=cos(pi*alltimepoints)/2+xx(:,2)+(1+0.25*sin(3*pi*alltimepoints))*rts(ntimepoints)
!xx(:,3)=-sin(pi*alltimepoints)/2+xx(:,3)+(1+0.25*cos(3*pi*alltimepoints))*rts(ntimepoints)
!end function

!!!!!!!!!!!!!!!
!!  Model 3  !!
!!!!!!!!!!!!!!!

!function random_gamma42() !n=shape,r=rate
!double precision::temp(4),random_gamma42
!call random_number(temp)
!random_gamma42=-sum(log(temp))/2
!if(random_gamma42>20)random_gamma42=20
!if(random_gamma42<-20)random_gamma42=-20
!end function

!function random_t10()
!double precision::temp(10),random_t10,num,den
!temp=rnorm(10)
!num=sum(temp)/10.0
!den=sqrt(sum((temp-num)**2)/90.0)
!random_t10=num/den
!if(random_t10>20)random_t10=20
!if(random_t10<-20)random_t10=-20
!end function

!function generate_xx() result(xx)
!double precision::xx(ntimepoints,dimbeta),temp(ntimepoints)
!logical::temp_logic(ntimepoints)
!integer::tt

!xx(:,1)=0.3*random_t10()
!xx(:,2)=0.3*random_gamma42()
!do tt=1,ntimepoints
! xx(tt,1)=xx(tt,1)+random_t10()*(1-0.5*alltimepoints(tt)**2)
! xx(tt,2)=xx(tt,2)+random_t10()*(1+0.25*sin(3*pi*alltimepoints(tt)))
!end do

!call random_number(temp)
!temp_logic=temp<0.001
!xx(1,3)=0
!do tt=2,ntimepoints
! if(temp_logic(tt))then
!  xx(tt,3)=xx(tt-1,3)+1
! else
!  xx(tt,3)=xx(tt-1,3)
! end if
!end do
!xx(:,3)=0.2*xx(:,3)+rnorm(ntimepoints)*(1+0.25*cos(3*pi*alltimepoints(tt)))

!xx(:,2)=xx(:,2)-0.5*sin(2*pi*alltimepoints)
!xx(:,3)=xx(:,3)-0.5*cos(2*pi*alltimepoints)
!end function

! SUBROUTINE 3 - ESTIMATION

function estimate_beta(hh_int)
integer::mm,ii,ll,jj,hh_int
double precision,dimension(dimbeta)::estimate_beta,beta_old,beta_new
double precision::U(dimbeta),I(dimbeta,dimbeta)
double precision::denominator,numerator_vec(dimbeta),numerator_mat(dimbeta,dimbeta)
double precision::Kexp,Kexpxxij(dimbeta)
double precision::frac_vec(dimbeta),frac_mat(dimbeta,dimbeta)
double precision,dimension(-ntimepoints:ntimepoints)::allkvalues
double precision::COFACTOR(dimbeta,dimbeta),DETERMINANT,SOLVEIU(dimbeta)

hh=hh_int*omega
allkvalues=0d0
do jj=-hh_int,hh_int
 allkvalues(jj)=0.75*(1-(dble(jj)/hh_int)**2)/hh
end do

print*,"aa"

beta_old=beta_true;beta_new=beta_true
do mm=1,100
 U=0;I=0
 do ii=1,nind
  if(.NOT.delta(ii))cycle
  U=U+xxij(ii,mmi(ii),:)
  denominator=0;numerator_vec=0;numerator_mat=0
  do ll=1,nind
   if(ot_int(ll)<ot_int(ii))cycle
   do jj=1,mmi(ll)
    if(abs(ot_int(ii)-ttij_int(ll,jj))>=hh_int)cycle
    Kexp=allkvalues(ot_int(ii)-ttij_int(ll,jj))*exp(dot_product(xxij(ll,jj,:),beta_old))
    denominator=denominator+Kexp
    Kexpxxij=Kexp*xxij(ll,jj,:)
    numerator_vec=numerator_vec+Kexp*xxij(ll,jj,:)
    
    numerator_mat(1,1)=numerator_mat(1,1)+Kexpxxij(1)*xxij(ll,jj,1)
    numerator_mat(2,1)=numerator_mat(2,1)+Kexpxxij(2)*xxij(ll,jj,1)
    numerator_mat(3,1)=numerator_mat(3,1)+Kexpxxij(3)*xxij(ll,jj,1)
    numerator_mat(2,2)=numerator_mat(2,2)+Kexpxxij(2)*xxij(ll,jj,2)
    numerator_mat(3,2)=numerator_mat(3,2)+Kexpxxij(3)*xxij(ll,jj,2)
    numerator_mat(3,3)=numerator_mat(3,3)+Kexpxxij(3)*xxij(ll,jj,3)
    
   end do
  end do
  
  numerator_mat(1,2)=numerator_mat(2,1)
  numerator_mat(1,3)=numerator_mat(3,1)
  numerator_mat(2,3)=numerator_mat(3,2)
  
  frac_vec=numerator_vec/denominator
  frac_mat=numerator_mat/denominator
  U=U-frac_vec
  I=I-frac_mat+spread(frac_vec,2,dimbeta)*spread(frac_vec,1,dimbeta)
 end do
 
   COFACTOR(1,1) = I(2,2)*I(3,3)-I(2,3)*I(3,2)
   !COFACTOR(1,2) = -(I(2,1)*I(3,3)-I(2,3)*I(3,1))
   !COFACTOR(1,3) = +(I(2,1)*I(3,2)-I(2,2)*I(3,1))
   COFACTOR(2,1) = -I(1,2)*I(3,3)+I(1,3)*I(3,2)
   COFACTOR(2,2) = I(1,1)*I(3,3)-I(1,3)*I(3,1)
   !COFACTOR(2,3) = -(I(1,1)*I(3,2)-I(1,2)*I(3,1))
   COFACTOR(3,1) = I(1,2)*I(2,3)-I(1,3)*I(2,2)
   COFACTOR(3,2) = -I(1,1)*I(2,3)+I(1,3)*I(2,1)
   COFACTOR(3,3) = I(1,1)*I(2,2)-I(1,2)*I(2,1)
   
   DETERMINANT=I(3,1)*COFACTOR(3,1)+I(3,2)*COFACTOR(3,2)+I(3,3)*COFACTOR(3,3)
   !SOLVEIU=matmul(COFACTOR,U)/DETERMINANT
   
   SOLVEIU(1)=(COFACTOR(1,1)*U(1)+COFACTOR(2,1)*U(2)+COFACTOR(3,1)*U(3))/DETERMINANT
   SOLVEIU(2)=(COFACTOR(2,1)*U(1)+COFACTOR(2,2)*U(2)+COFACTOR(3,2)*U(3))/DETERMINANT
   SOLVEIU(3)=(COFACTOR(3,1)*U(1)+COFACTOR(3,2)*U(2)+COFACTOR(3,3)*U(3))/DETERMINANT

   beta_new=beta_old-SOLVEIU
   
 if(sum(abs(beta_new-beta_old))<condition_value)exit
 beta_old=beta_new
end do
estimate_beta=beta_new
end function estimate_beta


function estimate_mu(hh_int)
double precision::U(2,2),V(2),estimate_mu(ntimepoints),hh
integer,intent(in)::hh_int
integer::tt,ii,jj
double precision,dimension(-ntimepoints:ntimepoints)::allkvalues,allktvalues,allkttvalues

hh=hh_int*omega
allkvalues=0d0;allktvalues=0d0;allkttvalues=0d0
do jj=-hh_int,hh_int
 allkvalues(jj)=0.75*(1-(dble(jj)/hh_int)**2)/hh
 allktvalues(jj)=allkvalues(jj)*jj*omega
 allkttvalues(jj)=allktvalues(jj)*jj*omega
end do

do tt=1,ntimepoints
 U=0;V=0
 do ii=1,nind
  if(ot_int(ii)<tt)cycle
  do jj=1,mmi(ii)
   if(abs(tt-ttij_int(ii,jj))>=hh_int)cycle
   U(1,1)=U(1,1)+allkvalues(ttij_int(ii,jj)-tt)
   U(1,2)=U(1,2)+allktvalues(ttij_int(ii,jj)-tt)
   U(2,2)=U(2,2)+allkttvalues(ttij_int(ii,jj)-tt)
   V(1)=V(1)+allkvalues(ttij_int(ii,jj)-tt)*rrij(ii,jj)
   V(2)=V(2)+allktvalues(ttij_int(ii,jj)-tt)*rrij(ii,jj)
  end do
 end do
 U(2,1)=U(1,2)
 estimate_mu(tt)=(U(2,2)*V(1)-U(1,2)*V(2))/(U(2,2)*U(1,1)-U(1,2)*U(2,1))
end do
end function estimate_mu


function estimate_sigma2(hh_int)
double precision::U(2,2),V(2),estimate_sigma2(ntimepoints)
integer,intent(in)::hh_int
integer::tt,ii,jj
double precision,dimension(-ntimepoints:ntimepoints)::allkvalues,allktvalues,allkttvalues

hh=hh_int*omega
allkvalues=0d0;allktvalues=0d0;allkttvalues=0d0
do jj=-hh_int,hh_int
 allkvalues(jj)=0.75*(1-(dble(jj)/hh_int)**2)/hh
 allktvalues(jj)=allkvalues(jj)*jj*omega
 allkttvalues(jj)=allktvalues(jj)*jj*omega
end do

do tt=1,ntimepoints
 U=0;V=0
 do ii=1,nind
  if(ot_int(ii)<tt)cycle
  do jj=1,mmi(ii)
   if(abs(tt-ttij_int(ii,jj))>=hh_int)cycle
   U(1,1)=U(1,1)+allkvalues(ttij_int(ii,jj)-tt)
   U(1,2)=U(1,2)+allktvalues(ttij_int(ii,jj)-tt)
   U(2,2)=U(2,2)+allkttvalues(ttij_int(ii,jj)-tt)
   V(1)=V(1)+allkvalues(ttij_int(ii,jj)-tt)*epsij2(ii,jj)
   V(2)=V(2)+allktvalues(ttij_int(ii,jj)-tt)*epsij2(ii,jj)
  end do
 end do
 U(2,1)=U(1,2)
 estimate_sigma2(tt)=(U(2,2)*V(1)-U(1,2)*V(2))/(U(2,2)*U(1,1)-U(1,2)*U(2,1))
end do
end function estimate_sigma2

! SUBROUTINE 5 - TUNING PARAMETERS

function calculate_CVL(hh_int)
double precision,dimension(dimbeta)::beta_old,beta_new,beta_leave
integer::iileave,mm,ii,jj,kk,ll
double precision::lliscore,calculate_CVL(nhh_beta)
double precision::U(dimbeta),I(dimbeta,dimbeta)
double precision::numerator,denominator,numerator_vec(dimbeta),numerator_mat(dimbeta,dimbeta)
double precision::Kexp,Kexp_vec(dimbeta),Kexp_mat(dimbeta,dimbeta)
double precision::frac_vec(dimbeta),frac_mat(dimbeta,dimbeta)
double precision::sumK,sumKexp,sumK_leave,sumKexp_leave
double precision::Kexpxxij(dimbeta)
integer,parameter::hh_beta_int_list(nhh_beta)=(/(ii*40,ii=1,nhh_beta)/)
integer::hh_index,hh_int
double precision::hh
double precision,dimension(-ntimepoints:ntimepoints)::allkvalues
double precision::COFACTOR(dimbeta,dimbeta),DETERMINANT,SOLVEIU(dimbeta)

calculate_CVL=0
do hh_index=1,nhh_beta
 hh_int=hh_beta_int_list(hh_index)
 hh=hh_int*omega
 allkvalues=0d0
 do jj=-hh_int,hh_int
  allkvalues(jj)=0.75*(1-(dble(jj)/hh_int)**2)/hh
 end do
 do iileave=1,nind !,20
  beta_old=beta_true;beta_new=beta_true
  do mm=1,100
   U=0;I=0
   !$OMP PARALLEL DEFAULT(SHARED),&
   !$OMP &PRIVATE(ii,jj,kk,ll,denominator,numerator_vec,numerator_mat),&
   !$OMP &PRIVATE(Kexp,Kexpxxij,frac_vec,frac_mat),&
   !$OMP &REDUCTION(+:U,I)
   !$OMP DO
   do ii=1,nind
    if((.NOT.delta(ii)).OR.iileave==ii)cycle
    U=U+xxij(ii,mmi(ii),:)
    denominator=0
    numerator_vec=0
    numerator_mat=0
    do ll=1,nind
     if(iileave==ii)cycle
     if(ot_int(ll)<ot_int(ii))exit
     do jj=mmi(ll),1,-1
      if(ttij_int(ll,jj)-ot_int(ii)>=hh_int)cycle
      if(ttij_int(ll,jj)-ot_int(ii)<=-hh_int)exit
      Kexp=allkvalues(ot_int(ii)-ttij_int(ll,jj))*exp(dot_product(xxij(ll,jj,:),beta_old))
      denominator=denominator+Kexp
      Kexpxxij=Kexp*xxij(ll,jj,:)
      numerator_vec=numerator_vec+Kexp*xxij(ll,jj,:)
      
      numerator_mat(1,1)=numerator_mat(1,1)+Kexpxxij(1)*xxij(ll,jj,1)
      numerator_mat(2,1)=numerator_mat(2,1)+Kexpxxij(2)*xxij(ll,jj,1)
      numerator_mat(3,1)=numerator_mat(3,1)+Kexpxxij(3)*xxij(ll,jj,1)
      numerator_mat(2,2)=numerator_mat(2,2)+Kexpxxij(2)*xxij(ll,jj,2)
      numerator_mat(3,2)=numerator_mat(3,2)+Kexpxxij(3)*xxij(ll,jj,2)
      numerator_mat(3,3)=numerator_mat(3,3)+Kexpxxij(3)*xxij(ll,jj,3)
      
     end do
    end do
    
    numerator_mat(1,2)=numerator_mat(2,1)
    numerator_mat(1,3)=numerator_mat(3,1)
    numerator_mat(2,3)=numerator_mat(3,2)
    
    frac_vec=numerator_vec/denominator
    frac_mat=numerator_mat/denominator
    U=U-frac_vec
    I=I-frac_mat+spread(frac_vec,2,dimbeta)*spread(frac_vec,1,dimbeta)
   end do
   !$OMP END PARALLEL

   COFACTOR(1,1) = I(2,2)*I(3,3)-I(2,3)*I(3,2)
   !COFACTOR(1,2) = -(I(2,1)*I(3,3)-I(2,3)*I(3,1))
   !COFACTOR(1,3) = +(I(2,1)*I(3,2)-I(2,2)*I(3,1))
   COFACTOR(2,1) = -I(1,2)*I(3,3)+I(1,3)*I(3,2)
   COFACTOR(2,2) = I(1,1)*I(3,3)-I(1,3)*I(3,1)
   !COFACTOR(2,3) = -(I(1,1)*I(3,2)-I(1,2)*I(3,1))
   COFACTOR(3,1) = I(1,2)*I(2,3)-I(1,3)*I(2,2)
   COFACTOR(3,2) = -I(1,1)*I(2,3)+I(1,3)*I(2,1)
   COFACTOR(3,3) = I(1,1)*I(2,2)-I(1,2)*I(2,1)
   
   DETERMINANT=I(3,1)*COFACTOR(3,1)+I(3,2)*COFACTOR(3,2)+I(3,3)*COFACTOR(3,3)
   !SOLVEIU=matmul(COFACTOR,U)/DETERMINANT
   
   SOLVEIU(1)=(COFACTOR(1,1)*U(1)+COFACTOR(2,1)*U(2)+COFACTOR(3,1)*U(3))/DETERMINANT
   SOLVEIU(2)=(COFACTOR(2,1)*U(1)+COFACTOR(2,2)*U(2)+COFACTOR(3,2)*U(3))/DETERMINANT
   SOLVEIU(3)=(COFACTOR(3,1)*U(1)+COFACTOR(3,2)*U(2)+COFACTOR(3,3)*U(3))/DETERMINANT

   beta_new=beta_old-SOLVEIU
   
   if(sum(abs(beta_new-beta_old))<condition_value)exit
   beta_old=beta_new
  end do
  beta_leave=beta_new

  if(delta(iileave))then
   lliscore=0
  else
   lliscore=1
  end if
  do ii=nind,1,-1
   if(ot_int(ii)>ot_int(iileave))exit
   if(.not.delta(ii))cycle
   numerator=0;denominator=0
   do jj=mmi(iileave),1,-1
    if(ttij_int(iileave,jj)-ot_int(ii)>=hh_int)cycle
    if(ttij_int(iileave,jj)-ot_int(ii)<=-hh_int)exit
    numerator=numerator+allkvalues(ot_int(ii)-ttij_int(iileave,jj))*exp(dot_product(xxij(iileave,jj,:),beta_leave))
   end do
   do ll=nind,1,-1
    if(ot_int(ll)>ot_int(ii))exit
    do jj=mmi(ll),1,-1
     if(ttij_int(ll,jj)-ot_int(ii)>=hh_int)cycle
     if(ttij_int(ll,jj)-ot_int(ii)<=-hh_int)exit
     denominator=denominator+allkvalues(ot_int(ii)-ttij_int(ll,jj))*exp(dot_product(xxij(ll,jj,:),beta_leave))
    end do
   end do
   lliscore=lliscore-numerator/denominator
  end do
  calculate_CVL(hh_index)=calculate_CVL(hh_index)+lliscore**2
 end do
end do

hh_index=minloc(calculate_CVL,dim=1)
hh_int=hh_beta_int_list(hh_index)
end function


function calculate_CVmu_faster(hh_int) result(CVmu)
double precision::U(2,2),V(2),mu_leave(nmax_obs),RSSmui,CVmu(nhh_mu)
integer::tt,iileave,jjleave,ii,jj,hh_index,hh_int
integer,parameter::hh_mu_int_list(nhh_mu)=(/(ii*20,ii=1,nhh_mu)/)
double precision::hh
double precision,dimension(-ntimepoints:ntimepoints)::allkvalues,allktvalues,allkttvalues

CVmu=0d0
do hh_index=1,nhh_mu
 hh_int=hh_mu_int_list(hh_index)
 hh=hh_int*omega
 allkvalues=0d0;allktvalues=0d0;allkttvalues=0d0
 do jj=-hh_int,hh_int
  allkvalues(jj)=0.75*(1-(dble(jj)/hh_int)**2)/hh
  allktvalues(jj)=allkvalues(jj)*jj*omega
  allkttvalues(jj)=allktvalues(jj)*jj*omega
 end do
 !$OMP PARALLEL DEFAULT(SHARED),&
 !$OMP &PRIVATE(tt,ii,jj,iileave,jjleave,U,V,RSSmui,mu_leave),&
 !$OMP &REDUCTION(+:CVmu)
 !$OMP DO
 do iileave=1,nind
  do jjleave=1,mmi(iileave)
   tt=ttij_int(iileave,jjleave)
   U=0d0;V=0d0
   do ii=1,nind
    if(ii==iileave)cycle
    if(ot_int(ii)<tt)exit
    do jj=1,mmi(ii)
     if(abs(tt-ttij_int(ii,jj))>=hh_int)cycle
     U(1,1)=U(1,1)+allkvalues(ttij_int(ii,jj)-tt)
     U(1,2)=U(1,2)+allktvalues(ttij_int(ii,jj)-tt)
     U(2,2)=U(2,2)+allkttvalues(ttij_int(ii,jj)-tt)
     V(1)=V(1)+allkvalues(ttij_int(ii,jj)-tt)*rrij(ii,jj)
     V(2)=V(2)+allktvalues(ttij_int(ii,jj)-tt)*rrij(ii,jj)
    end do
   end do
   U(2,1)=U(1,2)
   mu_leave(jjleave)=(U(2,2)*V(1)-U(1,2)*V(2))/(U(2,2)*U(1,1)-U(1,2)*U(2,1))
  end do
  RSSmui=sum((rrij(iileave,1:mmi(iileave))-mu_leave(1:mmi(iileave)))**2)
  CVmu(hh_index)=CVmu(hh_index)+RSSmui
 end do
 !$OMP END PARALLEL
end do
hh_index=minloc(CVmu,dim=1)
hh_int=hh_mu_int_list(hh_index)
end function calculate_CVmu_faster


function calculate_CVsigma2_faster(hh_int) result(CVsigma2)
double precision::U(2,2),V(2),sigma2_leave(nmax_obs),RSSsigma2i,CVsigma2(nhh_sigma2)
integer::tt,iileave,jjleave,ii,jj,hh_index,hh_int
integer,parameter::hh_sigma2_int_list(nhh_sigma2)=(/(ii*30,ii=1,nhh_sigma2)/)
double precision::hh
double precision,dimension(-ntimepoints:ntimepoints)::allkvalues,allktvalues,allkttvalues

CVsigma2=0d0
do hh_index=1,nhh_sigma2
 hh_int=hh_sigma2_int_list(hh_index)
 hh=hh_int*omega
 allkvalues=0d0;allktvalues=0d0;allkttvalues=0d0
 do jj=-hh_int,hh_int
  allkvalues(jj)=0.75*(1-(dble(jj)/hh_int)**2)/hh
  allktvalues(jj)=allkvalues(jj)*jj*omega
  allkttvalues(jj)=allktvalues(jj)*jj*omega
 end do
 !$OMP PARALLEL DEFAULT(SHARED),&
 !$OMP &PRIVATE(tt,ii,jj,iileave,jjleave,U,V,RSSsigma2i,sigma2_leave),&
 !$OMP &REDUCTION(+:CVsigma2)
 !$OMP DO
 do iileave=1,nind
  do jjleave=1,mmi(iileave)
   tt=ttij_int(iileave,jjleave)
   U=0d0;V=0d0
   do ii=1,nind
    if(ii==iileave)cycle
    if(ot_int(ii)<tt)exit
    do jj=1,mmi(ii)
     if(abs(tt-ttij_int(ii,jj))>=hh_int)cycle
     U(1,1)=U(1,1)+allkvalues(ttij_int(ii,jj)-tt)
     U(1,2)=U(1,2)+allktvalues(ttij_int(ii,jj)-tt)
     U(2,2)=U(2,2)+allkttvalues(ttij_int(ii,jj)-tt)
     V(1)=V(1)+allkvalues(ttij_int(ii,jj)-tt)*epsij2(ii,jj)
     V(2)=V(2)+allktvalues(ttij_int(ii,jj)-tt)*epsij2(ii,jj)
    end do
   end do
   U(2,1)=U(1,2)
   sigma2_leave(jjleave)=(U(2,2)*V(1)-U(1,2)*V(2))/(U(2,2)*U(1,1)-U(1,2)*U(2,1))
  end do
  RSSsigma2i=sum((epsij2(iileave,1:mmi(iileave))-sigma2_leave(1:mmi(iileave)))**2)
  CVsigma2(hh_index)=CVsigma2(hh_index)+RSSsigma2i
 end do
 !$OMP END PARALLEL
end do
hh_index=minloc(CVsigma2,dim=1)
hh_int=hh_sigma2_int_list(hh_index)
end function calculate_CVsigma2_faster

subroutine sort_by_time
integer::idx,temp_int,temp_int_arr(nmax_obs)
logical::temp_logic
double precision::temp_db_arr(nmax_obs,dimbeta)

do ii=1,(nind-1)
 idx=maxloc(ot_int(ii:nind),dim=1)+ii-1
 
 !! ot_int
 temp_int=ot_int(ii)
 ot_int(ii)=ot_int(idx)
 ot_int(idx)=temp_int
 
 !! mmi
 temp_int=mmi(ii)
 mmi(ii)=mmi(idx)
 mmi(idx)=temp_int
 
 !! delta
 temp_logic=delta(ii)
 delta(ii)=delta(idx)
 delta(idx)=temp_logic
 
 !! ttij
 temp_int_arr=ttij_int(ii,:)
 ttij_int(ii,:)=ttij_int(idx,:)
 ttij_int(idx,:)=temp_int_arr
 
 !! xxij
 temp_db_arr=xxij(ii,:,:)
 xxij(ii,:,:)=xxij(idx,:,:)
 xxij(idx,:,:)=temp_db_arr
end do

end subroutine

function choose_ll(lambda) result(ll)
integer,parameter::nchosen=50000
double precision::unif01(nchosen),variance,numerator,EWMA
integer::chosen(nchosen),ii,jj,cond_s,cond_n
double precision::ll,ll_left,ll_right,ATS
logical::signal(nind)

double precision,intent(in)::lambda
double precision::temp
integer::accuracy

 call random_number(unif01)
 chosen=floor(unif01*nind)+1
 ll_right=0.5
 do while(ll_right<2.5)
  cond_s=0;cond_n=0
  do ii=1,nchosen
   variance=1-(1-lambda)**21
   EWMA=variance*(rrij_std(chosen(ii),1)+random_normal()*0.001)
   if(EWMA>ll_right)then
    cond_s=cond_s+ttij_int(chosen(ii),1)
    cond_n=cond_n+1
    cycle
   end if
   if(mmi(chosen(ii))==1)cycle
   do jj=2,mmi(chosen(ii))
    temp=(1-lambda)**(ttij_int(chosen(ii),jj)-ttij_int(chosen(ii),jj-1))
    variance=variance/(variance+temp)
    EWMA=(1-variance)*EWMA+variance*(rrij_std(chosen(ii),jj)+random_normal()*0.001)
    if(EWMA>ll_right)then
     cond_s=cond_s+ttij_int(chosen(ii),jj)
     cond_n=cond_n+1
     exit
    end if
   end do
  end do
  ATS=cond_s/dble(cond_n)
  if(ATS>370)exit
  ll_right=ll_right+0.03
 end do

 ll_left=ll_right-0.03
 do while(ll_right-ll_left>0.0001)
  ll=(ll_left+ll_right)/2
  cond_s=0;cond_n=0
  do ii=1,nchosen
   variance=1-(1-lambda)**21
   EWMA=variance*(rrij_std(chosen(ii),1)+random_normal()*0.001)
  
   if(EWMA>ll)then
    cond_s=cond_s+ttij_int(chosen(ii),1)
    cond_n=cond_n+1
    cycle
   end if
   if(mmi(chosen(ii))==1)cycle
   do jj=2,mmi(chosen(ii))
    temp=(1-lambda)**(ttij_int(chosen(ii),jj)-ttij_int(chosen(ii),jj-1))
    variance=variance/(variance+temp)
    EWMA=(1-variance)*EWMA+variance*(rrij_std(chosen(ii),jj)+random_normal()*0.001)
    
    if(EWMA>ll)then
     cond_s=cond_s+ttij_int(chosen(ii),jj)
     cond_n=cond_n+1
     exit
    end if
   end do
  end do
  ATS=cond_s/dble(cond_n)
  if(ATS<370)then
   ll_left=ll
  else
   ll_right=ll
  end if
 end do

end function

function choose_ll_d() result(ll)
integer,parameter::nchosen=50000,nlambda=4
double precision,parameter::lambda(nlambda)=[0.01,0.015,0.02,0.05]
double precision::unif01(nchosen),variance(nlambda),numerator(nlambda),EWMA(nlambda),temp(nlambda)
integer::chosen(nchosen),ii,jj,cond_s,cond_n
double precision::ll,ll_left,ll_right,ATS
logical::signal(nind)

 call random_number(unif01)
 chosen=floor(unif01*nind)+1
 ll_right=2.0
 do while(ll_right<3.6)
  cond_s=0;cond_n=0
  do ii=1,nchosen
   variance=1
   numerator=rrij_std(chosen(ii),1)+random_normal()*0.001
   EWMA=numerator
  
   if(maxval(EWMA)>ll_right)then
    cond_s=cond_s+ttij_int(chosen(ii),1)
    cond_n=cond_n+1
    cycle
   end if
   if(mmi(chosen(ii))==1)cycle
   do jj=2,mmi(chosen(ii))
    temp=(1-lambda)**(ttij_int(chosen(ii),jj)-ttij_int(chosen(ii),jj-1))
    variance=1+variance*temp**2
    numerator=rrij_std(chosen(ii),jj)+random_normal()*0.001+numerator*temp
    EWMA=numerator/sqrt(variance)
    if(maxval(EWMA)>ll_right)then
     cond_s=cond_s+ttij_int(chosen(ii),jj)
     cond_n=cond_n+1
     exit
    end if
   end do
  end do
  ATS=cond_s/dble(cond_n)
  if(ATS>370)exit
  ll_right=ll_right+0.03
 end do
 
 ll_left=ll_right-0.03
 do while(ll_right-ll_left>0.0001)
  ll=(ll_left+ll_right)/2
  cond_s=0;cond_n=0
  do ii=1,nchosen
   variance=1
   numerator=rrij_std(chosen(ii),1)+random_normal()*0.001
   EWMA=numerator
  
   if(maxval(EWMA)>ll)then
    cond_s=cond_s+ttij_int(chosen(ii),1)
    cond_n=cond_n+1
    cycle
   end if
   if(mmi(chosen(ii))==1)cycle
   do jj=2,mmi(chosen(ii))
    temp=(1-lambda)**(ttij_int(chosen(ii),jj)-ttij_int(chosen(ii),jj-1))
    variance=1+variance*temp**2
    numerator=rrij_std(chosen(ii),jj)+random_normal()*0.001+numerator*temp
    EWMA=numerator/sqrt(variance)
    if(maxval(EWMA)>ll)then
     cond_s=cond_s+ttij_int(chosen(ii),jj)
     cond_n=cond_n+1
     exit
    end if
   end do
  end do
  ATS=cond_s/dble(cond_n)
  if(ATS<370)then
   ll_left=ll
  else
   ll_right=ll
  end if
 end do
 
end function

subroutine monitor(ll,lambda,ATS,table)
double precision,intent(in)::ll,lambda
double precision,intent(out)::ATS
integer,intent(out)::table(4)
double precision::variance,numerator,EWMA
integer::ii,jj,cond_s,cond_n
logical::signal(nind_OC)
double precision::temp

 cond_s=0;cond_n=0
 signal=.FALSE.
 do ii=1,nind_OC
  variance=1-(1-lambda)**21
  EWMA=variance*rrij_std_OC(ii,1)
  if(EWMA>ll)then
   signal(ii)=.TRUE.
   cond_s=cond_s+ttij_int_OC(ii,1)
   cond_n=cond_n+1
   cycle
  end if
  if(mmi_OC(ii)==1)cycle
  do jj=2,mmi_OC(ii)
   temp=(1-lambda)**(ttij_int_OC(ii,jj)-ttij_int_OC(ii,jj-1))
   variance=variance/(variance+temp)
   EWMA=(1-variance)*EWMA+variance*rrij_std_OC(ii,jj)

   if(EWMA>ll)then
    signal(ii)=.TRUE.
    cond_s=cond_s+ttij_int_OC(ii,jj)!!
    cond_n=cond_n+1!!
    exit
   end if
  end do
 end do
 table(1)=count(signal.and.delta_OC)
 table(2)=count(signal.and.(.not.delta_OC))
 table(3)=count((.not.signal).and.delta_OC)
 table(4)=count((.not.signal).and.(.not.delta_OC))
 ATS=cond_s/dble(cond_n)
end subroutine


subroutine monitor_d(ll,ATS,table)
integer,parameter::nlambda=4
double precision,parameter::lambda(nlambda)=[0.01,0.015,0.02,0.05]
double precision,intent(in)::ll
double precision,intent(out)::ATS
integer,intent(out)::table(4)
double precision::variance(nlambda),numerator(nlambda),EWMA(nlambda)
integer::ii,jj,cond_s,cond_n
logical::signal(nind_OC)
double precision::temp(nlambda)

 cond_s=0;cond_n=0
 signal=.FALSE.
 do ii=1,nind_OC
  variance=1
  numerator=rrij_std_OC(ii,1)
  EWMA=numerator
  if(maxval(EWMA)>ll)then
   signal(ii)=.TRUE.
   cond_s=cond_s+ttij_int_OC(ii,1)
   cond_n=cond_n+1
   cycle
  end if
  if(mmi_OC(ii)==1)cycle
  do jj=2,mmi_OC(ii)
   temp=(1-lambda)**(ttij_int_OC(ii,jj)-ttij_int_OC(ii,jj-1))
   variance=1+variance*temp**2
   numerator=rrij_std_OC(ii,jj)+numerator*temp
   EWMA=numerator/sqrt(variance)
   if(maxval(EWMA)>ll)then
    signal(ii)=.TRUE.
    cond_s=cond_s+ttij_int_OC(ii,jj)
    cond_n=cond_n+1
    exit
   end if
  end do
 end do
 table(1)=count(signal.and.delta_OC)
 table(2)=count(signal.and.(.not.delta_OC))
 table(3)=count((.not.signal).and.delta_OC)
 table(4)=count((.not.signal).and.(.not.delta_OC))
 ATS=cond_s/dble(cond_n)
end subroutine


end program main_program
